@@ -350,13 +350,46 @@ namespace xt
     {
     };
 
-#define XTENSOR_FORWARD_METHOD(name)           \
-    auto name() const                          \
-        -> decltype(std::declval<E>().name())  \
-    {                                          \
-        return m_ptr->name();                  \
+#define XTENSOR_FORWARD_CONST_METHOD(name)                                      \
+    auto name() const                                                           \
+        -> decltype(std::declval<xtl::constify_t<E>>().name())                  \
+    {                                                                           \
+        return m_ptr->name();                                                   \
     }
 
+#define XTENSOR_FORWARD_METHOD(name)                                            \
+    auto name() -> decltype(std::declval<E>().name())                           \
+    {                                                                           \
+        return m_ptr->name();                                                   \
+    }
+
+#define XTENSOR_FORWARD_CONST_ITERATOR_METHOD(name)                             \
+    template <layout_type L = XTENSOR_DEFAULT_TRAVERSAL>                        \
+    auto name() const noexcept                                                  \
+        -> decltype(std::declval<xtl::constify_t<E>>().template name<L>())      \
+    {                                                                           \
+        return m_ptr->template name<L>();                                       \
+    }                                                                           \
+    template <layout_type L = XTENSOR_DEFAULT_TRAVERSAL, class S>               \
+    auto name(const S& shape) const noexcept                                    \
+        -> decltype(std::declval<xtl::constify_t<E>>().template name<L>(shape)) \
+    {                                                                           \
+        return m_ptr->template name<L>();                                       \
+    }
+
+#define XTENSOR_FORWARD_ITERATOR_METHOD(name)                                   \
+    template <layout_type L = XTENSOR_DEFAULT_TRAVERSAL, class S>               \
+    auto name(const S& shape) noexcept                                          \
+        -> decltype(std::declval<E>().template name<L>(shape))                  \
+    {                                                                           \
+        return m_ptr->template name<L>();                                       \
+    }                                                                           \
+    template <layout_type L = XTENSOR_DEFAULT_TRAVERSAL>                        \
+    auto name() noexcept -> decltype(std::declval<E>().template name<L>())      \
+    {                                                                           \
+        return m_ptr->template name<L>();                                       \
+    }                                                                           \
+
     namespace detail
     {
         template <class E>
@@ -469,16 +502,38 @@ namespace xt
             return m_ptr->operator()(args...);
         }
 
-        XTENSOR_FORWARD_METHOD(shape);
-        XTENSOR_FORWARD_METHOD(dimension);
-        XTENSOR_FORWARD_METHOD(size);
-        XTENSOR_FORWARD_METHOD(begin);
-        XTENSOR_FORWARD_METHOD(cbegin);
-        XTENSOR_FORWARD_METHOD(storage_begin);
-        XTENSOR_FORWARD_METHOD(storage_cbegin);
-        XTENSOR_FORWARD_METHOD(storage_end);
-        XTENSOR_FORWARD_METHOD(storage_cend);
-        XTENSOR_FORWARD_METHOD(layout);
+        XTENSOR_FORWARD_CONST_METHOD(shape)
+        XTENSOR_FORWARD_CONST_METHOD(dimension)
+        XTENSOR_FORWARD_CONST_METHOD(size)
+        XTENSOR_FORWARD_CONST_METHOD(layout)
+
+        XTENSOR_FORWARD_ITERATOR_METHOD(begin)
+        XTENSOR_FORWARD_ITERATOR_METHOD(end)
+        XTENSOR_FORWARD_CONST_ITERATOR_METHOD(begin)
+        XTENSOR_FORWARD_CONST_ITERATOR_METHOD(end)
+        XTENSOR_FORWARD_CONST_ITERATOR_METHOD(cbegin)
+        XTENSOR_FORWARD_CONST_ITERATOR_METHOD(cend)
+
+        XTENSOR_FORWARD_ITERATOR_METHOD(rbegin)
+        XTENSOR_FORWARD_ITERATOR_METHOD(rend)
+        XTENSOR_FORWARD_CONST_ITERATOR_METHOD(rbegin)
+        XTENSOR_FORWARD_CONST_ITERATOR_METHOD(rend)
+        XTENSOR_FORWARD_CONST_ITERATOR_METHOD(crbegin)
+        XTENSOR_FORWARD_CONST_ITERATOR_METHOD(crend)
+        
+        XTENSOR_FORWARD_METHOD(storage_begin)
+        XTENSOR_FORWARD_METHOD(storage_end)
+        XTENSOR_FORWARD_CONST_METHOD(storage_begin)
+        XTENSOR_FORWARD_CONST_METHOD(storage_end)
+        XTENSOR_FORWARD_CONST_METHOD(storage_cbegin)
+        XTENSOR_FORWARD_CONST_METHOD(storage_cend)
+        
+        XTENSOR_FORWARD_METHOD(storage_rbegin)
+        XTENSOR_FORWARD_METHOD(storage_rend)
+        XTENSOR_FORWARD_CONST_METHOD(storage_rbegin)
+        XTENSOR_FORWARD_CONST_METHOD(storage_rend)
+        XTENSOR_FORWARD_CONST_METHOD(storage_crbegin)
+        XTENSOR_FORWARD_CONST_METHOD(storage_crend)
 
         template <class T = E>
         std::enable_if_t<has_strides<T>::value, const inner_strides_type&>
@@ -530,12 +585,12 @@ namespace xt
         }
 
         template <class It>
-        auto element(It first, It last) {
+        reference element(It first, It last) {
             return m_ptr->element(first, last);
         }
 
         template <class It>
-        auto element(It first, It last) const {
+        const_reference element(It first, It last) const {
             return m_ptr->element(first, last);
         }
 
